# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import sys

import database_connection as dc

"""
Creates the SQLite database and its associated tables
"""


# This database holds all the Nodes and all the swicharoo packages and the fingers pointing on our packages
def create_software_database():
    if dc.database_exists():
        print('The database already exists.')
        exit(-1)

    print('Creating database:')

    connection = dc.get_connection(False)
    if connection is not None:
        print('Creating tables:')
        
        # Table for keeping track of other nodes: (unix time is seconds since 1970, 1jan 00:00:00)
        # our_ip and our_port is what the node think is our address, (like our_ip_according_to_this_node but that name is kind of long)
        # this is used when we change IP and/or PORT, so that we know we went off the radar on them.
        # The scheduler will figure that out and send our new IP:PORT to them.
        connection.execute('''CREATE TABLE node
                 (id INTEGER PRIMARY KEY AUTOINCREMENT,
                 ip         CHAR,
                 port       CHAR,
                 pubkey     CHAR,
                 connection_tries INTEGER DEFAULT 0,
                 connection_fails INTEGER DEFAULT 0,
                 last_connection_unixtime INTEGER DEFAULT 0,
                 our_ip         CHAR DEFAULT 'unknown',
                 our_port       CHAR DEFAULT 'unknown',
                 rem        CHAR
                  );''')
                  
        # Create two tables that will hold the data for fingers.
        # These tables will hold all our files we share, plus all files we share fo other users.

        # Create a table that will hold data we want to share
        # data_version is 1 for the original data, when refreshing, increment with 1 so fingers get stale and scheduler will send new data to sharing nodes
        # address_translation_file is for, you know it! Used so that we can update them automatically when our, or any other concerned node address changes
        #  should point on another data_owned file (so we know how to deal with stuff)  ?? Right or wrong?
        connection.execute('''CREATE TABLE data_owned
                 (id INTEGER PRIMARY KEY AUTOINCREMENT,
                 name               CHAR,
                 address_translation_file CHAR DEFAULT 0,
                 originalfilename   CHAR,
                 key                CHAR,
                 iv                 CHAR,
                 pubkey             CHAR,
                 data_version INTEGER DEFAULT 1,
                 wanted_fingers     INTEGER DEFAULT 10,
                 address_list       CHAR,
                 address_list_epoc_time INTEGER DEFAULT 0,
                 unique (name)
                  );''')

        # Create a table that holds data that other nodes wants us to share
        connection.execute('''CREATE TABLE data_shared
                 (id INTEGER PRIMARY KEY AUTOINCREMENT,
                 name               CHAR,
                 pubkey             CHAR,
                 data_version INTEGER DEFAULT 1,
                 unique (name, pubkey)
                  );''')

        # Table for keeping track of who is keeping and sharing my files, points on 'node'(the node keeping the data) & 'data_owned' (my data)
        # Typically we would like to have 10 entries for each entry in the finger table that points on a local file (our file) in 'data'
        # data_version is for checking the data_owned is not stale
        connection.execute('''CREATE TABLE finger
                 (id INTEGER PRIMARY KEY AUTOINCREMENT,
                 nodeid     INTEGER,
                 dataid     INTEGER,
                 data_version INTEGER DEFAULT 1,
                 connection_tries INTEGER DEFAULT 0,
                 connection_fails INTEGER DEFAULT 0,
                 last_connection_unixtime INTEGER DEFAULT 0
                  );''')

        # Address list for shared data
        # Each entry is an address for one data stored in a node
        # A node is referenced by pubkey, ip, port
        # The data is referenced by dataid
        # The data is validated by the owners privkey, stored in the digest
        # Our data (that we share) is not (yet) in here (? why)
        # The 'encrypted' field contains a json representation of the address, encrypted with the encryption the data uses
        # We should remove all the fields except 'dataid' and 'encrypted_address' (we can keep 'id' for convenience)
        # TODO: There is a lot of crap in here, like the encrypted address stuff
        connection.execute('''CREATE TABLE address
                 (id INTEGER PRIMARY KEY AUTOINCREMENT,
                 pubkey     CHAR,
                 ip         CHAR,
                 port       CHAR,
                 dataid     INTEGER,
                 digest     CHAR,
                 encrypted_address CHAR
                  );''')


        # When we store data for another node, the data name is not unique (say "Music", "text" etc.)
        # so we store them in a folder unique for each node (e.g the index in this table)
        # Here we can find the name of the folder (the id) from a nodes public key
        connection.execute('''CREATE TABLE folder
                 (folderid INTEGER PRIMARY KEY AUTOINCREMENT,
                 pubkey     CHAR
                  );''')

        # My RSA keypair (just one entry, for now)
        connection.execute('''CREATE TABLE rsa
                 (id INTEGER PRIMARY KEY AUTOINCREMENT,
                 privkey    CHAR,
                 pubkey     CHAR
                  );''')

        # Configuration data
        connection.execute('''CREATE TABLE configuration_string
                 (name    CHAR,
                 value     CHAR
                  );''')

        connection.execute('''CREATE TABLE configuration_int
                 (name    CHAR,
                 value     INTEGER
                  );''')


        print('Database created.')

        set_database_default_values()
        print('Default configuration values inserted, edit with -s.')

    else:
        print('There were errors.')


def set_database_default_values():
    if not dc.database_exists():
        print('The database does not yet exists.')
        return

    connection = dc.get_connection(False)
    if connection is not None:
        # Insert default values:
        cursor = connection.cursor()

        # Truncate tables:
        cursor.execute("DELETE FROM configuration_string")
        cursor.execute("DELETE FROM configuration_int")

        cursor.execute("INSERT INTO configuration_string ('name', 'value') VALUES ('ip', '127.0.0.1')")
        cursor.execute("INSERT INTO configuration_string ('name', 'value') VALUES ('unsecure_nick_name', 'Alice')")
        cursor.execute("INSERT INTO configuration_string ('name', 'value') VALUES ('swicharoo_policy', 'GRANT_ALL')")

        # Exception to the rule: We can have several of these exclude regex:
        cursor.execute("INSERT INTO configuration_string ('name', 'value') VALUES ('exclude_ip_list', 'nothing')")

        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('port', 46000)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('listen_read_bytes', 10000)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('max_simultaneous_connections', 5)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('listen_timeout_seconds', 5)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('server_send_package_size', 32768)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('max_storage_space_other_mb', 16384)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('swicharoo_minsize_accept_all', 65536)")
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('swicharoo_file_maxsize_kb', 65536)")
        # How long time we wait between the different scheduler jobs
        cursor.execute("INSERT INTO configuration_int ('name', 'value') VALUES ('scheduler_wait_msec', 1000)")

        connection.commit()
    else:
        print('There were errors.')


if len(sys.argv) > 1 and sys.argv[1] == '--generate':
    import database_query as dq
    if not dc.database_exists():
        create_software_database()
