# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import sys, os
import json
import database_query as dq
import rsa_helper as rh
from configuration import get_configuration_value
import sqlite3
from sqlite3 import Error
import address_registry as ar


"""
Creates a link file from existing data (that was created with the insert_data.py)
"""

def print_help():
    print('Usage:')
    print('')
    print('python3 get_link_from_data.py dataname')
    print('python3 get_link_from_data.py --testrun')

    sys.exit()


# Returns the connection so we don't need to open it again, or None if it fails
def generate_database(database_file_name):
    # delete any old db
    if os.path.isfile(database_file_name):
        os.remove(database_file_name)

    # Generate the database
    try:
        connection = sqlite3.connect(database_file_name)
    except Error:
        print("Could not create the (database) link:")
        print(Error)
        return None

    connection.execute('''CREATE TABLE information
             (
              name                  TEXT,
              originalfilename      TEXT,
              file_version          INTEGER DEFAULT 1,
              key                   TEXT,
              iv                    INTEGER,
              data_rsa_pub          TEXT
               );''')

    # The addresses where we'll try to fetch the data
    # address_list.append({'rsa_pub': pub, 'ip': ip, 'port': port})
    connection.execute('''CREATE TABLE address_list
             (
              rsa_pub       TEXT,
              ip            TEXT,
              port          TEXT
             );''')

    return connection


def set_information_in_database(connection, name, original_file_name, version, aes_key, aes_iv, data_rsa_pub_key):
    cursor = connection.cursor()
    cursor.execute('INSERT INTO information (name, originalfilename, file_version, key, iv, data_rsa_pub) VALUES(?, ?, ?, ?, ?, ?)', (name, original_file_name, version, aes_key, aes_iv, data_rsa_pub_key))
    connection.commit()


def add_address_to_database(connection, rsa_pub_key, ip, port):
    cursor = connection.cursor()
    cursor.execute('INSERT INTO address_list (rsa_pub, ip, port) VALUES(?, ?, ?)', (rsa_pub_key, ip, port))
    connection.commit()


# Takes 'Name' entry from global database and generates a link named 'name'.10f which contains
# This link is a SQLite database and contains two tables;
# One for global information: IP, PORT, NAME, ORIGFILENAME, AES&IV, RSA_PUB
# (AES is stored in 64 bytes hex string)
# One address entry for each finger storing the data we are generating the link for.
# If version == 0 then the latest version for this data will be used
# if exclude_our_address == True then we will not add our own address to the link, only other nodes links
# Fingers are all the fingers which addresses we shall add to the .10f database
def _create_and_save_link_database(name, exclude_our_address, fingers):
    row = dq.get_link_data_by_name(name)
    key_pair = dq.get_rsa_hex_keys()

    if row is None or key_pair is None:
        if row is None:
            print('#Error: Link not found in datbase:', name)
        if key_pair is None:
            print('#Error: RSA keys not found in datbase.')
        return

    # Get data id from global database:
    data_id = dq.get_link_dataid_by_name(name)

    # get the latest data_version
    data_version = dq.get_owned_data_version(data_id)

    connection = generate_database(name + '.10f')
    if connection is None:
        return

    prv, pub = key_pair

    # Update the database link with 'my' address information
    if not exclude_our_address:
        add_address_to_database(connection, pub, get_configuration_value('ip'), get_configuration_value('port'))

    print("Fingers:", len(fingers), flush=True)

    fingers_len = len(fingers)
    for n in range(0, fingers_len):
        finger_id = fingers[n]
        # add to the address list:
        node_id,_ = dq.get_finger_data(finger_id)
        if node_id != 0:
            # get rsa_pub, ip, port from node and add it to the address list
            node_data = dq.get_node_data(node_id)
            # Update the link with address information
            add_address_to_database(connection, node_data[2], node_data[0], node_data[1])

    # data_rsa is, together with the name, the unique identifier for this data (e.g the rsa_pub key from the node owning the data)
    _, data_rsa_pub = rh.get_rsa_keypair(True)
    data_rsa_pub = data_rsa_pub.exportKey('DER').hex()

    # Update the link with information
    set_information_in_database(connection, name, row[1], data_version, row[2], row[3], data_rsa_pub)

    print('Link created:', name + '.10f')


# We now generate both files
def create_and_save_link_database(name, exclude_our_address=False):

    # Get data id from global database:
    data_id = dq.get_link_dataid_by_name(name)
    translation_data_id = dq.get_link_dataid_by_name(ar.generate_trans_name(name))
    print("data id:", data_id)
    print("translation data id:", translation_data_id)

    # Get a list of finger ids for the data we want to share and for the addresslist we want to share
    data_fingers = dq.get_all_fingers(data_id)
    translation_fingers = dq.get_all_fingers(translation_data_id)
    if len(data_fingers) == 0:
        print('Warning: This data have not yet been uploaded elsewhere! If your PC is shut down, your link will not function')
    if len(translation_fingers) == 0:
        print('Warning: Translation address list has not been uploaded elsewhere. Your link will expire if nodes change addresses.')

    print("Nodes sharing the data:", data_fingers)
    print("Nodes sharing the address translation file:", translation_fingers)

    # Make the data link
    _create_and_save_link_database(name, exclude_our_address, data_fingers)
    # Make the address translator link, which contains the list of addresses used in the data link
    if len(translation_fingers) > 0:
        _create_and_save_link_database(ar.generate_trans_name(name), exclude_our_address, translation_fingers)
