# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import os
import sys
from sys import exit
import threading
import signal
import socket
import logging
import time
from Cryptodome.PublicKey import RSA
import datetime
from struct import pack, unpack
import json
import random

from configuration import get_configuration_value
from socket_communication import SecureCommunicationListen
import database_query as dq
import updater as ud
import rsa_helper as rh
import locator
from helpers import human_readable_file_size
from accept_swicharoo import auto_accept_small_file_for_swicharoo, storage_space_enough_for_swicharoo

from ip_version import typeOfIPAddress


"""
the listener serves encrypted data to anyone knowing the name of the data
It stores data in a SQLite database
"""

# Set global flogging:
logging.basicConfig(
     filename='listener.log',
     level=logging.INFO,
     format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
     datefmt='%H:%M:%S'
)
# Set up logging to console
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')  # set simple console format
console.setFormatter(formatter)
logging.getLogger('').addHandler(console)  # add the handler to the root logger

logger = logging.getLogger(__name__)
# Logging setup done

# Global shut down boolean
force_shutdown = False


class Listen:
    def __init__(self):
        self.listensocket = None
        self.clientsocket = None

        print('Allocated storage space for other nodes:' + human_readable_file_size(get_configuration_value('max_storage_space_other_mb')*1024*1024))
        print('Max filesize for other nodes:' + human_readable_file_size(get_configuration_value('swicharoo_file_maxsize_kb')*1024))
        print('Always accept files under this size (if we have storage space):' + human_readable_file_size(get_configuration_value('swicharoo_minsize_accept_all')))

    def forcestop(self):
        # logging.info("Shutting down listener...")
        if self.listensocket is not None:
            self.listensocket.close()

    def listenthread_secure(self, client_socket, client_address):
        # logging.info("Recv Message from" + str(client_address[0]))

        prv, pub = dq.get_rsa_hex_keys()
        # logging.info("Recv Message, make secure link:")
        scl = SecureCommunicationListen(client_socket, RSA.importKey(bytes.fromhex(prv)), get_configuration_value('listen_read_bytes'))
        if not scl.is_active():
            # logging.info('Listener could not setup secure link with caller, bailing out...')
            return
        # print('Secure link created')
        n = scl.recv()

        # logging.info("Recv Message done.")
        if n:
            # logging.info('Received msg, size=' + str(len(n)))
            pass
        else:
            # logging.info('Recved nothing, close socket and bail out.')
            return
        # logging.info('10-part:' + str(n[:10]))

        # All requests take up 10 bytes
        request = n[:10]
        data = n[10:]
        # print('REQ:', request)
        # print('MSG:', data)
        try:
            # What does he want?

            if request == b'AREUALIVE?':  # Just check if we are a Tenfingers node
                # logging.info('Listen thread got AREUALIVE req.')
                scl.send(b'YES')

            # This is the download data request
            if request == b'ONEFINGER:':  # Rest of the request is a name + RSA_PubKey. If the name corresponds in the database, send back the encrypted file.
                # logging.info('Listen thread got ONEFINGER req.')
                data = data.decode()
                datalist = data.split(',')  # todo: This should just be a name, don't forget to escape it!

                if len(datalist) == 2:
                    # print('datalist:', datalist)
                    name = datalist[0]
                    rsa_pub = datalist[1]
                    # logging.info('Request for: [' + name + ']')

                    # TODO, Serve the name or an error!

                    row = dq.get_link_data_by_name_and_rsapub(name, rsa_pub)

                    if row is not None:

                        # logging.info('Found [' + name + '], points on file [' + row[0] + ']')
                        # Get data from DB
                        file = row[0]
                        our_version = row[1]
                        data = None

                        # Send back version
                        print('Node wants:', file)
                        print('Propose version', our_version)
                        answer = scl.send(str(our_version).encode())

                        if answer == b'GO':

                            # get the folder that is in the /data/ folder (based on the table 'folder' / or our own local folder 'local'):
                            filefolder = os.path.join(locator.create_path('data/') + dq.get_folder(rsa_pub))

                            # logging.info('Trying to serve from folder:' + filefolder)

                            with open(filefolder + "/" + file + '.aes', 'rb') as f:
                                data = f.read()
                            if data is None:
                                # logging.info('[' + name + '] does not point on a file or it was not readable.')
                                scl.send(b'FileNotFound')
                            else:
                                # logging.info('[' + name + '] corresponds to a file and we loaded it, lets send it back:')

                                # Send it back (maybe we need to convert the data, it is raw bytes now)
                                scl.send(data)
                        else:
                            # They bailed
                            print('Our data is too old')
                    else:
                        # logging.info('[' + name + '] not found in database.' + rsa_pub)
                        scl.send(b'NotFound')
                else:
                    # logging.info('Malformed request.')
                    scl.send(b'BadRequest')

            # Update request

            if request == b'UPDATEDATA':  # another node asks us to update data we store for them
                logging.info('Listen thread got UPDATEDATA req.')

                filesize = unpack('>Q', data[:8])[0]  # todo: always use data[:8], so a ',' in the pack(...) won't break things! SO fix it everywhere
                # print('Update data REQ, filesize', filesize)

                arr = data[8:].split(b',')  # todo: always use data[8:], so a ',' in the pack(...) won't break things! SO fix it everywhere

                if len(arr) == 5:
                    other_name = arr[1].decode()
                    other_pub_hex_bytes = arr[2]
                    other_pub_hex = other_pub_hex_bytes.decode()
                    new_version = arr[3].decode()

                    # print('Verify connecting node...')

                    # First verify the caller is actually the owner of this public rsa key:
                    random_checksum = b'Digest_this!'+str(random.randint(10000, 99999)).encode()+str(time.time()).encode()
                    # print('Send random verification data:', random_checksum)
                    verification_digest = scl.send(random_checksum)  # caller should now have made a digest off of this checksum
                    # print('Got verification digest:', verification_digest)

                    other_pub_key = RSA.importKey(bytes.fromhex(other_pub_hex))
                    # print('Key type:', type(other_pub_key))

                    # print('Check digest:')
                    digest_ok = rh.verify_digest(random_checksum, other_pub_key, verification_digest)

                    if digest_ok:
                        # print('Digest OK, lets check data size:', filesize)
                        # print('other_name type:', type(other_name))
                        logging.info('Got update request for file '+other_name+' of size:'+str(filesize))

                        success = True
                        if filesize > get_configuration_value('swicharoo_file_maxsize_kb')*1024:
                            logging.info('Update denied, file too big. Size:' + str(filesize))
                            scl.send(b'Too Big!')
                            success = False

                        # Check the version is a number
                        if not new_version:
                            new_version = 1
                        if new_version.isdigit() is False:
                            logging.info('Update denied, version is not a number')
                            scl.send(b'Version not a number.')
                            success = False

                        if success:
                            logging.info('Update to version ' + str(new_version))
                            # Check we have the place to store the file:
                            used_space = dq.get_shared_files_size()
                            if used_space + filesize > get_configuration_value('max_storage_space_other_mb')*1024*1024:
                                logging.info('Update denied, we are out of disc space.')
                                scl.send(b'Out of storage space, sorry.')
                                success = False

                        # TODO: Check the swicharoo doesn't break the swicharoo rules according to the original shared file!
                        # TODO Or else a malicous user can swicharoo a small file, then update it to a much bigger file.
                        if success:
                            pass
                            # Check our swicharoo policy!

                        if success:
                            # print('File size OK.')

                            # Verify we actually store this file for this node!
                            shared_data_id = dq.get_shared_data_id_by_name_and_rsapub(other_name, other_pub_hex)
                            if shared_data_id != 0:
                                # print('Update request: all good!')

                                # Check the version is actually a newer one (we know the entry exists in the database)
                                old_version = dq.get_shared_data_version(shared_data_id)
                                if int(old_version) < int(new_version):

                                    # logging.info('Update accepted, starting the update.')
                                    new_file_data = scl.send(b'ACCEPTING!')

                                    # Save other nodes data
                                    filefolder = os.path.join(locator.create_path('data/'), dq.get_folder(other_pub_hex))
                                    if not os.path.exists(filefolder):
                                        os.makedirs(filefolder)

                                    save_name = os.path.join(filefolder, other_name + '.aes')
                                    # logging.info('Save off other file data: ' + save_name)
                                    with open(save_name, 'wb') as f:
                                        f.write(new_file_data)

                                    # Success, update to this new version!
                                    dq.set_shared_data_version(shared_data_id, new_version)

                                    # logging.info('Update done (success).')
                                    logging.info('Update succeeded.')
                                    scl.send(b'OK')
                                else:
                                    logging.info('Update not needed, we are up to date.')
                                    scl.send(b'NO_NEED_TO')

                            else:
                                # print('Update data: we do not store that data here')
                                logging.info('Update failed, we do not store your data (any more)')
                                scl.send(b'DataRemoved')
                        else:
                            logging.info('Update failed.')
                            scl.send(b'Updatefailed')
                    else:
                        logging.info('Update failed, the digest was wrong.')
                        scl.send(b'UNDIGESTABLE')
                else:
                    logging.info('Update denied, malformed request.')
                    scl.send(b'Malformed update request! len=' + str(len(arr)))

                # The classic swicharoo:
            if request == b'SWICHAROO?':  # The good old swicharoo

                # We just got a request for swicharoo with a file size and a pub key
                data_size = data[:8]  # Get the first 8 bytes, who represents the others files size
                tmp = data[8:]  # Remove those first 8 bytes, the rest should be the other_pub_key,other_name,...
                split = tmp.decode().split('|')
                if len(split) != 5:
                    print("#Listener Error in split size, expected five got", len(split))
                    client_socket.close()
                    logging.info('Listener: Got badly formed SWICHAROO string')
                    return

                # We now have all the information about the switch, from their side.
                other_pub_key = split[0]
                other_name = split[1]
                other_ip = split[2]
                other_port = split[3]
                other_data_version = split[4]
                other_file_size = unpack('>Q', data_size)[0]
                logging.info('Listener: Got a Swicharoo request, other_file_size: ' + str(other_file_size) + ' file name: ' + other_name)

                # Check the version exists and is a number
                if not other_data_version:
                    other_data_version = 1
                if other_data_version.isdigit() is False:
                    logging.info('Listener: Swicharoo request turned down because version is not a number')
                    scl.send(b'VERSIONNAN')
                    client_socket.close()
                    return

                # Find the data we want to send back, but do not select data that the other node already store:
                dummy_data = False
                data_id = dq.get_valid_dataid_needing_finger_for_swicharoo(other_file_size, other_pub_key, other_name)
                if data_id <= 0:
                    if not storage_space_enough_for_swicharoo(other_file_size) or not auto_accept_small_file_for_swicharoo(other_file_size):
                        scl.send(b'SRYNODATA.')
                        logging.info('Listener: Swicharoo request turned down because we do not have anything to share that the offer corresponds to')
                        client_socket.close()
                        return
                    # Get or set up a universal small dummy share file
                    dummy_data = True

                if dummy_data:
                    # Make up some node-unique name here
                    my_name = 'Dummy_' + str(time.time())
                    my_origfilename = my_name
                    my_file_size = 21  # the 'LetsBeNiceToEachOther' below
                else:
                    # Get the name and info about our real file:
                    my_name, my_origfilename = dq.get_link_data_by_dataid(data_id)
                    my_file_size = os.path.getsize(os.path.join(locator.create_path('data/local/'), (my_name + '.aes')))

                print('#Listener try to share back data_id:', data_id, ' (0 = gracefully share their file)')

                random_checksum = b'Digest_this!' + str(random.randint(10000, 99999)).encode() + str(time.time()).encode()
                query = b'YES_PLEASE' + pack('>Q', my_file_size) + my_name.encode() + b'|' + my_origfilename.encode() + b'|' + random_checksum
                #   <<        Send Recv           >>
                logging.info('Listener: Send verification:')
                verification_digest = scl.send(query)
                logging.info('Listener: got back answer')

                # Check all the different things that have to work out if we should go further:

                if verification_digest == b'U_TOO_BIG!':
                    print('They declined our acceptance of their swicharoo request because our file is too big.', flush=True)
                    client_socket.close()
                    logging.info('Listener: Swicharoo turned down our file, too big')
                    return

                if verification_digest == b'DATA_AGAIN':
                    logging.info('Listener: They seems to already store our data.')
                    client_socket.close()
                    return

                digest_ok = rh.verify_digest(random_checksum, RSA.importKey(bytes.fromhex(other_pub_key)), verification_digest)
                if not digest_ok:
                    scl.send(b'DIGEST_ERR')
                    logging.info('Listener: Caller could not authentify themselves. Bail out.')
                    client_socket.close()
                    return
                logging.info('Listener: digest ok.')

                if dq.get_link_data_by_name_and_rsapub(other_name, other_pub_key) is not None:  # We already store the data
                    scl.send(b'ALREADY_DO')
                    logging.info('Listener: We already store their data.')
                    client_socket.close()
                    return

                # Check if we know this node, and if we don't, insert it into the database
                other_node_id = dq.get_nodeid(other_ip, other_port, other_pub_key)
                if other_node_id == 0:  # It didn't exist in our database, insert it:
                    logging.info('Listener: Insert new node')
                    dq.insert_node(other_ip, other_port, other_pub_key, 'From Listener SWICHAROO')
                    other_node_id = dq.get_nodeid(other_ip, other_port, other_pub_key)
                    dq.set_shared_data_version(other_node_id, other_data_version)

                # Load up our data so we can send it away:
                if dummy_data:
                    file_data = b'LetsBeNiceToEachOther'
                else:
                    with open(os.path.join(locator.create_path('data/local/'), (my_name + '.aes')), 'rb') as f:
                        file_data = f.read()
                if file_data is None:
                    scl.send(b'MISCONFIG.')
                    logging.info('Listener: Owned data is missing: data/local/' + my_name + '.aes')
                    scl.close()
                    return

                # All seems okay, proceed with the exchange and send them our data
                #   <<        Send Recv           >>
                other_data = scl.send(file_data)

                # Save other nodes data
                file_folder = os.path.join(locator.create_path('data/'), dq.get_folder(other_pub_key))
                if not os.path.exists(file_folder):
                    os.makedirs(file_folder)

                save_name = os.path.join(file_folder, other_name + '.aes')
                print('Listener: save theirs:', save_name, flush=True)
                logging.info('Listener: Save off other file data: ' + save_name)
                with open(save_name, 'wb') as f:
                    f.write(other_data)

                # I now have other: pubkey, name, origname, data, save off data and update DB
                dq.insert_data_shared(other_name, other_pub_key)  # Others data is now stored here for retrieval (by name + pubkey)

                # Insert a pointer to our data (their node)
                if not dummy_data:
                    logging.info('Insert finger')
                    dq.insert_finger(other_node_id, data_id)
                else:
                    logging.info('As we used dummy data, do not insert a finger but store in data_shared')

                logging.info('Done, send final OK.')
                # DONE!
                scl.send(b'OK')

            if request == b'SEND_NODES':  # send back a random node for now  # todo: send back several
                # check if we already have this node, and if not, add it:
                data = data.decode()
                datalist = data.split(',')  # This should be IP/PORT data from caller todo: Escape it!
                if len(datalist) == 3:
                    ip = datalist[0]
                    port = datalist[1]
                    rsa_pub = datalist[2]

                    # This call will silently fail if the node already exists, or it is our pubkey
                    dq.insert_node(ip, port, rsa_pub, 'From Listener SEND_NODES')

                else:
                    logging.info('#Warning: SEND_NODES req malformed, len(datalist) =' + str(len(datalist)))

                nodeid = dq.get_random_node_id(False, rsa_pub)  # todo: remove the False from the call so we send back only valid nodes
                if nodeid is not None:
                    tmp = dq.get_node_data(nodeid)
                    j = {'ip': tmp[0], 'port': tmp[1], 'rsa_pub_key': tmp[2]}
                    # print('Send back node:', tmp[0], tmp[1], tmp[2][:22], flush=True)
                    # print('j:', j)
                    try:
                        r = json.dumps(j)
                    except:
                        print('Could not dumps the dict')
                    # print('dumps=', r, flush=True)
                    b = r.encode()  # transform to byte object
                    # print('send:', b, flush=True)
                    scl.send(b)
                else:
                    scl.send(b'{}')
                    print('We do not have any nodes to share yet!', flush=True)
                # print('SEND NODES Done.', flush=True)

            if request == b'NEWADDRESS':  #
                # check if we already have this node, and if not, add it:
                data = data.decode()
                datalist = data.split(',')  # This should be IP/PORT data from caller todo: Escape it!
                if len(datalist) == 3:
                    print("NEWADDRESS: new request", flush=True)
                    other_ip = datalist[0]
                    other_port = datalist[1]
                    other_pub_key = datalist[2]
                    print("NEWADDRESS: other_ip=" + str(other_ip) + " other_port=" + str(other_port), flush=True)

                    # Cheap sanity check
                    if len(other_ip) < 64 and len(other_port) < 10:

                        # Update the node
                        print("NEWADDRESS: update", flush=True)
                        # Verify it is the right node (with their rsa pub + digest)

                        random_checksum = b'Digest_this!' + str(random.randint(10000, 99999)).encode() + str(time.time()).encode()
                        # query = b'YES_PLEASE' + pack('>Q', my_file_size) + my_name.encode() + b'|' + my_origfilename.encode() + b'|' + random_checksum
                        print("NEWADDRESS: send digest test and get back result", flush=True)
                        verification_digest = scl.send(random_checksum)
                        print("NEWADDRESS: check result", flush=True)
                        digest_ok = rh.verify_digest(random_checksum, RSA.importKey(bytes.fromhex(other_pub_key)), verification_digest)

                        if digest_ok:
                            print("NEWADDRESS: Digest ok", flush=True)

                            # We now know the caller is the rigtheous holder of the public key, so just update its address
                            print("NEWADDRESS: get node id", flush=True)
                            other_node_id = dq.get_nodeid_from_rsapubkey(other_pub_key)
                            if other_node_id:
                                print("NEWADDRESS: update", flush=True)
                                # Update the nodes address
                                dq.update_node_address(other_node_id, other_ip, other_port)
                                scl.send(b'OK')
                            else:
                                print("NEWADDRESS: found no node...", flush=True)
                                scl.send(b'Your address not found')
                        else:
                            print("NEWADDRESS: Digest wrong", flush=True)
                            scl.send(b'Digerst error')
                    else:
                        scl.send(b'Error NEWADDRESS request IP or PORT are not valid')
                else:
                    # Makformed request
                    scl.send(b'Error Malformed NEWADDRESS request')


            # Close the socket
            client_socket.close()

            if request == b'NEW_METADT':  # A node is requiring an update of its address list, metadata,...
                tmp = data.split(b',')  # name + ',' + pub + ',' + epoch_time + address_list, digest
                if len(tmp) == 5:
                    name = tmp[0]
                    pub = tmp[1]
                    epoch_time = tmp[2]
                    digest = tmp[3]
                    address_list = tmp[4]  # address list is not used in the digest.

                    # first we verify this public key shows the private key has signed the message:
                    t = name + ',' + pub + ',' + epoch_time
                    if rh.verify_digest(t.encode(), pub, digest):
                        # if so, then update the database (as data is stored by name+pub_key, this works)
                        print('NEW_METADT request: verification OK, insert meta data and address list')
                        dq.try_update_address_list(name, pub, address_list, epoch_time)
                    else:
                        print('NEW_METADT request: digest verification failed')
                else:
                    print('Got mal formed NEW_METADT request, incorrect size is:', len(tmp))

        except NameError as error:
            logging.info("[NameError] Message not understood: NameError=" + str(error))
        except:
            logging.info("[General Exception] Message not understood: " + str(sys.exc_info()[0]))
            scl.send(b'GeneralError')
        # logging.info('Listen thread done.')

        if scl.is_active():
            scl.close()

    def run(self):
        print('inner thread start')
        try:
            # Here we could/should(?) list all network interfaces and their supported protocols...
            ip = get_configuration_value('ip')
            ipversion = typeOfIPAddress(ip)
            if ipversion == 'Invalid':
                self.listensocket = None
                print('#Warning: tried to connect to an invalid ip address:', ip)
                sys.stdout.flush()
                return

            print('IP version detected:', ipversion, '(' + str(ip) + ')')
            sys.stdout.flush()
            # Connect to other node
            if ipversion == 'IPv4':
                self.listensocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if ipversion == 'IPv6':
                self.listensocket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM, 0)  # and/or socket.socket(family=socket.AF_INET6)

            # self.listensocket.bind((str(socket.INADDR_ANY), self.port))
            if ipversion == 'IPv4':
                self.listensocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.listensocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                self.listensocket.bind(('', get_configuration_value('port')))
            else:
                self.listensocket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
                self.listensocket.bind((ip, get_configuration_value('port')))

            self.listensocket.listen(get_configuration_value('max_simultaneous_connections'))
            self.listensocket.settimeout(get_configuration_value('listen_timeout_seconds'))

        except socket.error as err:
            print('Failed to create listen socket: ', err)
            logging.info('Failed to create listen socket')
            exit()

        # log
        logging.info('Listen on port: ' + str(get_configuration_value('port')))

        global force_shutdown
        while not force_shutdown:

            gotconnection = False
            while not force_shutdown and gotconnection is False:
                try:
                    # establish connection
                    client_socket, client_address = self.listensocket.accept()
                    # Pass through the while:
                    gotconnection = True
                    # Make a new thread that will deal with the communication:
                    if not force_shutdown:
                        # Read request:
                        client_socket.settimeout(60)
                        # Start listener
                        threading.Thread(target=self.listenthread_secure, args=(client_socket, client_address)).start()

                except socket.error:  # as err:
                    # print("listen exception", err)
                    pass


version = '1.1'

logging.info('Tenfingers listener startup: V' + version + ' start time: ' + str(datetime.datetime.now()))
logging.info('Use --noscheduler to disable scheduler')
logging.info('Listen configuration: Public IP=' + get_configuration_value('ip') + ' Listen port=' + str(get_configuration_value('port')))
logging.info("Scheduler CPU alleviation rest time (msec):" + str(get_configuration_value('scheduler_wait_msec')))

run_scheduler = True
if len(sys.argv) > 1 and sys.argv[1] == '--noscheduler':
    logging.info('Scheduler will not be started (--noscheduler)')
    run_scheduler = False


time.sleep(0.1) # So logging will have time to write

# set up a thread with the scheduler
from scheduler import Scheduler
scheduler = None
scheduler_thread = None
if run_scheduler:
    scheduler = Scheduler()
    scheduler_thread = threading.Thread(name='Scheduler', target=scheduler.run)
    scheduler_thread.start()
    time.sleep(0.1)  # So logging will have time to write

# Create thread and start up:
listener = Listen()
listener_thread = threading.Thread(name='Listener', target=listener.run)  # l.run2, args=("abc",))

logging.info("start listener thread:")
listener_thread.start()


# Set up a function that will be called at SIGINT so that a CTRL+C will shut down all threads and exit correctly:
def signal_handler(sig, frame):
        logging.info('Received SIGINT, shutting down threads.')
        global force_shutdown
        force_shutdown = True
        listener.forcestop()


# Set up the SIGINT handler
signal.signal(signal.SIGINT, signal_handler)
